# -*- coding: utf-8 -*-
from .fundosbr import *
from .benchmarks import *
from .comparador import *
from . import version

__version__ = version.__version__
__author__ = "Rafael Rodrigues, rafa-rod @ GitHub"